//
// Created by dawsh on 1/17/24.
//

#include <glad/glad.h>
#include <JGL/JGL.h>
#include <GL/glut.h>
#include "J3ML/LinearAlgebra/Transform2D.h"
#include <rewindow/types/window.h>
#include <freetype2/ft2build.h>
#include FT_FREETYPE_H
#include <iostream>


GLuint program;
GLuint texture;

namespace JGL
{
    Vector2 ScreenToViewport(const Vector2 &v) {
        // TODO: Implement (CORRECT!!!) matrix transformation

        Vector2 windowSize = ReWindow::RWindow::CurrentWindow()->getSize();

        //Transform2D transform;

        //transform = transform.Translate(1.f, 1.f); // Center
        //transform = transform.Scale(0.5f, 0.5f); // Scale by half
        //transform = transform.Scale(viewportWidth, viewportHeight); // Scale by screen size

        //return transform.Transform(v);

        float x = (v.x) / windowSize.x;
        float y = (v.y) / windowSize.y;

        return {
                x, y
        };
    }

    FT_Face face;
    FT_Library ft;

    using namespace J3ML;

    bool InitTextEngine() {
        constexpr u32 default_font_size = 16;
        if (FT_Init_FreeType(&ft))
        {
            std::cout << "Error::FREETYPE: " << std::endl;
            return false;
        }

        if (FT_New_Face(ft, "content/FreeSans.ttf", 0, &face))
        {
            std::cout << "Error::FREETYPE: Failed to load font!" << std::endl;
            return false;
        }
        FT_Set_Pixel_Sizes(face, 0, default_font_size);
        return true;
    }

    void RenderText(std::string text, float x, float y, float scale) {
        glUseProgram(0); // Fixed-function pipeline.
        glColor3f(1.0f, 1.0f, 1.0f);

        const char* c;
        for (c = text.c_str(); *c; c++)
        {
            if (FT_Load_Char(face, *c, FT_LOAD_RENDER))
                continue;

            FT_GlyphSlot g = face->glyph;

            glActiveTexture(GL_TEXTURE0);
            glBindTexture(GL_TEXTURE_2D, texture);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, g->bitmap.width, g->bitmap.rows, 0, GL_ALPHA, GL_UNSIGNED_BYTE, g->bitmap.buffer);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

            float x2 = x + g->bitmap_left * scale;
            float y2 = -y - g->bitmap_top * scale; // Adjust y-coordinate
            float w = g->bitmap.width * scale;
            float h = g->bitmap.rows * scale;

            glBegin(GL_TRIANGLES);

            glTexCoord2f(0, 0);
            glVertex2f(x2, y2);

            glTexCoord2f(0, 1);
            glVertex2f(x2, y2 + h);

            glTexCoord2f(1, 1);
            glVertex2f(x2 + w, y2 + h);

            glTexCoord2f(0, 0);
            glVertex2f(x2, y2);

            glTexCoord2f(1, 1);
            glVertex2f(x2 + w, y2 + h);

            glTexCoord2f(1, 0);
            glVertex2f(x2 + w, y2);

            glEnd();

            x += (g->advance.x >> 6) * scale;
            y += (g->advance.y >> 6) * scale;

        }

        glBindTexture(GL_TEXTURE_2D, 0); // Unbind texture
    }

    namespace J2D
    {
        void FillRect2D(const Color3 &color, const Vector2 &pos, const Vector2 &size) {
            auto vp_pos = ScreenToViewport(pos);
            auto vp_size = ScreenToViewport(size);
            glBegin(GL_QUADS);
            glColor3f(color.r/255.f, color.g/255.f, color.b/255.f);
            glVertex2f(vp_pos.x, vp_pos.y);
            glVertex2f(vp_pos.x, vp_pos.y + vp_size.y);
            glVertex2f(vp_pos.x + vp_size.x, vp_pos.y + vp_size.y);
            glVertex2f(vp_pos.x + vp_size.x, vp_pos.y);
            glEnd();
        }

        void OutlineRect2D(const Color3 &color, const Vector2 &pos, const Vector2 &size, float thickness) {
            auto vp_pos = ScreenToViewport(pos);
            auto vp_size = ScreenToViewport(size);
            glBegin(GL_LINE_LOOP);
            glLineWidth(thickness);
            glColor3f(color.r, color.g, color.b);
            glVertex2f(vp_pos.x, vp_pos.y);
            glVertex2f(vp_pos.x, vp_pos.y + vp_size.y);
            glVertex2f(vp_pos.x + vp_size.x, vp_pos.y + vp_size.y);
            glVertex2f(vp_pos.x + vp_size.x, vp_pos.y);
            glEnd();
        }

        void DrawLine2D(const Color3 &color, const Vector2 &A, const Vector2 &B, float thickness) {
            auto vp_a = ScreenToViewport(A);
            auto vp_b = ScreenToViewport(B);

            glBegin(GL_LINE);
            glLineWidth(thickness);
            glColor3f(color.r, color.g, color.b);
            glVertex2f(vp_a.x, vp_a.y);
            glVertex2f(vp_b.x, vp_b.y);
            glEnd();

        }

        void DrawLine2D(const Color3 &color, float x, float y, float w, float h, float thickness) {
            DrawLine2D(color, {x, y}, {w, h}, thickness);
        }

        void DrawPixel2D(const Color3 &color, float x, float y) {
            DrawPixel2D(color, {x, y});
        }

        void DrawPixel2D(const Color3 &color, const Vector2 &coordinates) {
            auto vp_pos = ScreenToViewport(coordinates);
            glBegin(GL_POINT);
            glColor3f(color.r, color.g, color.b);
            glVertex2f(vp_pos.x, vp_pos.y);
            glEnd();
        }

        void
        OutlineCircle2D(const Color3 &color, const Vector2 &center, float radius, int subdivisions, float thickness) {
            glBegin(GL_LINE_LOOP);
            GLfloat angle;
            glColor3f(color.r, color.g, color.b);
            float step = (2.f * M_PI) / (float)subdivisions;
            for (angle = 0.0f; angle < (2.f * M_PI); angle += step)
            {
                GLfloat x = radius * sin(angle);
                GLfloat y = radius * cos(angle);
                x += center.x;
                y += center.y;
                glVertex2f(x, y);
            }
            glEnd();
        }

        void FillCircle2D(const Color3 &color, const Vector2 &center, float radius, int subdivisions) {
            glBegin(GL_POLYGON);
            GLfloat angle;
            glColor3f(color.r, color.g, color.b);
            float step = (2.f * M_PI) / (float)subdivisions;
            for (angle = 0.0f; angle < (2.f * M_PI); angle += step)
            {
                GLfloat x = radius * sin(angle);
                GLfloat y = radius * cos(angle);
                x += center.x;
                y += center.y;
                glVertex2f(x, y);
            }
            glEnd();
        }
    }


    namespace J3D
    {
        void DrawLine3D(const Color3& color, const Vector3& A, const Vector3& B, float thickness)
        {
            glBegin(GL_LINE);
            glLineWidth(thickness);
            glColor3f(color.r, color.g, color.b);
            glVertex3f(A.x, A.y, A.z);
            glVertex3f(B.x, B.y, B.z);
            glEnd();
        }
    }
}